(define-module (alterator lookout evaluation)

  ;;std modules
  :use-module (srfi srfi-1)
  :use-module (srfi srfi-2)
  :use-module (srfi srfi-11)
  :use-module (srfi srfi-13)

  :use-module (ice-9 regex)
  :use-module (ice-9 pretty-print)

  ;;common extentions
  :use-module (alterator gettext)

  ;;common helper modules
  :use-module (alterator container)
  :use-module (alterator vector)
  :use-module (alterator object)
  :use-module (alterator gettext)
  :use-module (alterator str)
  :use-module (alterator plist)
  :use-module (alterator algo)
  :use-module (alterator woo)
  :use-module (alterator constraints)
  :use-module (alterator context)

  ;;simplest message passing
  :use-module (alterator mailbox)

  ;;for toplevel-begin
  :use-module (alterator security wrap)

  ;;presentation engines
  :use-module (alterator presentation common)
  :use-module (alterator presentation document)
  :use-module (alterator presentation container)
  :use-module (alterator presentation events)
  
  ;;session gate
  :use-module (alterator session common)
	     
  ;;lookout: need for correct eval
  :use-module (alterator lookout presentation)
  :use-module (alterator lookout widget)
  :use-module (alterator lookout root)
  :use-module (alterator lookout ident)
  :use-module (alterator lookout globals)
  :use-module (alterator lookout goto)

  :export (document:eval))

;; simple formatting

(define (bold txt)
  (string-append "<b>" txt "</b>"))

(define (small txt)
  (string-append "<small>" txt "</small>"))

;;; attributes
(define-macro (define-attribute <name> . <rest>)
  `(define ,<name> (make-attribute ',<name> ,@<rest>)))

;;common events
(define-attribute clicked)
(define-attribute double-clicked)
(define-attribute changed)
(define-attribute return-pressed)
(define-attribute selected)
(define-attribute toggled)

;;internal events
(define-attribute destroyed)
(define-attribute loaded)

;;general attributes
(define-attribute align)
(define-attribute pixmap)
(define-attribute focus)
(define-attribute widget-name)
(define-attribute max-width)
(define-attribute max-height)
(define-attribute tab-index)

;;text related attributes
(define-attribute alterability)
(define-attribute url)
(define-attribute echo)
(define-attribute text-wrap)
(define-attribute text)

;;layout related attributes
(define-attribute margin)
(define-attribute spacing)

;;progress attributes
(define-attribute value)
(define-attribute maximum)
(define-attribute minimum)
(define-attribute step)

;;tree/listbox/combobox attributes
(define-attribute current)
(define-attribute tree-items)
(define-attribute coordinates)
(define-attribute rows)
(define-attribute current-rows)
(define-attribute state-rows)
(define-attribute header)
(define-attribute active)
(define-attribute current-text)
(define-attribute expanded)

;;other attributes
(define-attribute help)
(define-attribute state)
(define-attribute tooltip)
(define-attribute title)

;;menubar/wizardface attributes
(define-attribute actions)
(define-attribute current-action)
(define-attribute action-activity 2)
(define-attribute steps)
(define-attribute action-remove)
(define-attribute action-add)
(define-attribute current-step)
(define-attribute steps-clear 0)
(define-attribute step-text 2)
(define-attribute step-pixmap 2)
(define-attribute action-text 2)
(define-attribute action-pixmap 2)

;;; init attributes
(define-macro (define-init-attribute <name>)
  `(define ,<name> (make-init-attribute ',<name>)))

(define-init-attribute parent)
(define-init-attribute type)
(define-init-attribute columns)
(define-init-attribute checked)
(define-init-attribute width)
(define-init-attribute height)
(define-init-attribute orientation)
(define-init-attribute colspan)
(define-init-attribute rowspan)

;;; basic widgets
(define-macro (define-widget <name> . <args>)
  `(define ,<name> (make-widget ,(symbol->string <name>) ,@<args>)))

(define-widget radio text)
(define-widget button text)
(define-widget label text)
(define-widget edit text)
(define-widget groupbox title)
(define-widget textbox text)
(define-widget tab-page text)
(define-widget checkbox text)
(define-widget box)
(define-widget spinbox)
(define-widget gridbox)
(define-widget tabbox)
(define-widget listbox)
(define-widget checklistbox)
(define-widget combobox)
(define-widget progressbar)
(define-widget slider)
(define-widget tree)
(define-widget menubar)
(define-widget spacer)
(define-widget separator)
(define-widget dateedit)
(define-widget timeedit)
(define-widget splitbox)

;;; backward compatibility
(define checklist checklistbox)

(define checklist-rows (make-meta-attribute
			    checklist-rows
			    ((meta-set self widget)
			     (let* ((v (cond-car (value-of self)))
				    (vlist (map (lambda(x) (if (pair? x) x (cons x #t))) v))
				    (slist (map cdr vlist)))
			       (simple-set widget (rows vlist))
			       (simple-set widget (state-rows slist))
			       (simple-notify-set widget "rows" vlist)
			       (simple-notify-set widget "state-rows" slist)))
			    ((meta-get self widget)
			     (let ((vlist (simple-get widget rows))
				   (slist (simple-get widget state-rows)))
			       (map (lambda(x y) (cons (car x) y))
				    vlist slist)))))

(define checklist-active-rows (make-meta-attribute
			          checklist-active-rows
			    ((meta-get self widget)
			     (let ((vlist (simple-get widget rows))
				   (slist (simple-get widget state-rows)))
			       (filter-map (lambda(x y) (and y (car x)))
				    vlist slist)))))


;;; meta-attributes
(define-attribute enum-rows)
(define enumref (make-meta-attribute enumref
				     ((meta-set self widget)
				      (and-let* ((url-data (cond-car (value-of self)))
				                 (url (if (pair? url-data) (car url-data) url-data))
				                 (url-args (if (pair? url-data) (cdr url-data) '()))
						 (variants (apply woo-list url url-args))
						 (t (simple-get widget type)))
						(simple-set widget
							    (enum-rows
							     (map (lambda(x) (woo-get-option x 'name))
								 variants)))
						(cond
						  ((member t '("listbox" "combobox"))
						   (let ((labels (map (lambda(x) (cons (woo-get-option x 'label) ""))
								      variants)))
						     (simple-set widget (rows labels))
						     (simple-notify-set widget "rows" labels)))
						  ((string=? t "checklistbox")
						   (let ((labels (map (lambda(x) (cons (woo-get-option x 'label) #f))
								      variants)))
						     (simple-set widget (rows labels))
						     (simple-notify-set widget "rows" labels))))))))

(define value (make-meta-attribute value
				   ((meta-set self widget)
				    (let ((v (cond-car (value-of self)))
					  (t (simple-get widget type))
					  (l (or (simple-get widget enum-rows) '())))
				      (cond
					((member t '("listbox" "combobox"))
					 (let* ((idx (or (string-list-index v l) -1)))
					   (simple-set widget (current idx))
					   (simple-notify-set widget "current" idx)))
					((string=? t "checklistbox")
					 (let* ((rlist (or (simple-get widget rows) '()))
						(vlist (string-cut (or v "") #\;))
						(slist (map (lambda(x) (->bool (member x vlist))) l)))
					   (simple-notify-set widget "state-rows" slist)
					   (simple-set widget (state-rows slist))
					   (simple-set widget (rows (map (lambda (x y) (cons (car x) y))
									 rlist slist)))))
					(else
					  (simple-set widget (value v))
					  (simple-notify-set widget "value" v)))))
				   ((meta-get self widget)
				    (let ((t (simple-get widget type))
					  (l (or (simple-get widget enum-rows) '())))
				      (cond
					((member t '("listbox" "combobox"))
					 (let ((c (or (simple-get widget current) -1)))
					   (if (>= c 0) (list-ref l c) "")))
					((string=? t "checklistbox")
					 (let ((slist (or (simple-get widget state-rows) '())))
					   (string-join (filter-map (lambda(x y) (and x y)) slist l)
							";")))
					(else (simple-get widget value)))))))

;;; DOM
(define namelist (make-hidden-attribute 'namelist))

(define name (make-meta-attribute name
                   ((meta-set self widget)
                     (let* ((v (cond-car (value-of self)))
                            (r (global 'document:root))
                            (nlist (or (simple-get r namelist) '())))
                     (simple-set r (namelist (acons v widget nlist)))
                     (simple-set widget (name v))
                     (simple-notify-set widget "name" v)))))

(define (document-get-elements-by-name document name)
  (filter-map (lambda (x) (and (string=? (car x) name) (cdr x))) (document namelist)))

;;; i18n
(define-macro (with-translation <prefix> <textdomain> <instruction> . <instructions>)
          `(let ((,<prefix> (make-translator
                             ,<textdomain>
			     (session-language) )))
                ,<instruction> . ,<instructions>))

;;;;;;;;;;;;;;;

(define real-current-module (current-module))

(define (document:error . args) (error "try to use insecure function"))

;shield insecure functions like set!
(define (shield-insecure lst)
  (let ((insecure-vars '(set! load use-modules open-output-file
 			  open-file delete-file
			  eval with-output-to-file call-with-output-file
			  module-obarray-set!)))
    `(let ,(map (lambda (x) `(,x document:error)) insecure-vars) ,@lst)))

(define (document:eval url)
  (eval  (shield-ids
	   (shield-insecure
	     (preprocess-file url))) real-current-module))
